//
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
//
//
// win32.cpp
//
// win32 system class
//
// $Id: $
//

#include <iostream>
#include <time.h>
#include <sys/types.h>
#include "SDL.h"
#include "win32.h"
#include "event.h"
#include "config.h"
#include "version"

#define  DEF_VID_WIDTH        320
#define  DEF_VID_HEIGHT       240
#define  DEF_VID_BPP          32
#define  DEF_VID_SCALE        1
#define  DEF_VID_FULLSCREEN   0

#define  DEF_AUD_ENABLED      1
#define  DEF_AUD_SAMPLERATE   44100


System *System_factory()
{
   return new win32;
}


win32::win32()
: System()
{
   mpScreen = NULL;

   mWidth = DEF_VID_WIDTH;
   mHeight = DEF_VID_HEIGHT;
   mBPP = DEF_VID_BPP;
   mScale = DEF_VID_SCALE;
   mFullScreen = DEF_VID_FULLSCREEN;

   mSoundEnabled = true;
   mSampleRate = 0;

   // read config settings
   if (mpConfig)
   {
      mWidth = mpConfig->read_int("win32video", "width", mWidth);
      mHeight = mpConfig->read_int("win32video", "height", mHeight);
      mBPP = mpConfig->read_int("win32video", "bpp", mBPP);
      mScale = mpConfig->read_int("win32video", "scale", mScale);
      mFullScreen = mpConfig->read_int("win32video", "fullscreen", mFullScreen);

      mSoundEnabled = mpConfig->read_int("win32audio", "enabled", mSoundEnabled);
      mSampleRate = mpConfig->read_int("win32audio", "samplerate", mSampleRate);
   }

   if (SDL_Init(SDL_INIT_VIDEO))
      exit(-1);

   if (openVideo())
      exit(-1);

   openAudio();
}


win32::~win32()
{
   closeAudio();
   closeVideo();

   SDL_Quit();
}


void win32::openAudio()
{
   if (!mSoundEnabled)
      return;

   // set to false for now
   mSoundEnabled = false;

   // TODO: check failure case
   SDL_InitSubSystem(SDL_INIT_AUDIO);
   SDL_AudioSpec params;
   params.format = AUDIO_U8;
   params.channels = 1;
   params.freq = mSampleRate;
   params.callback = audioCallbackDispatch;
   params.userdata = (void *) this;

   // find closest power of two samples > frag size
   int i = 1;
   while (i < mSampleRate / 60)
      i <<= 1;

   params.samples = i;

   if (SDL_OpenAudio(&params, 0) < 0)
      return;

   int buffer_size = 2 * params.samples;
   mpAudioStream = new unsigned char[buffer_size];
   if (NULL == mpAudioStream)
      return;

   mSoundEnabled = true;
   mAudioFragSize = buffer_size;
   mSampleRate = params.freq;

   silenceAudio();
   SDL_PauseAudio(0);
}


void win32::audioCallbackDispatch(void *pUserData, Uint8 *pStream, int numSamples)
{
   ((win32 *) pUserData)->audioCallback((unsigned char *) pStream, numSamples);
}


void win32::closeAudio()
{
   if (mSoundEnabled)
   {
      SDL_CloseAudio();
      delete [] mpAudioStream;
   }
}


int win32::openVideo()
{
   int flags;

   flags = SDL_HWSURFACE | SDL_HWPALETTE;// | SDL_ANYFORMAT;
   if (mFullScreen)
      flags |= SDL_FULLSCREEN;

   mpScreen = SDL_SetVideoMode(mWidth, mHeight, mBPP, flags);
   if (NULL == mpScreen)
      return -1;

   SDL_ShowCursor(0);
   SDL_WM_SetCaption(APPNAME, NULL);

   lockVideo();

   mBPP = mpScreen->format->BitsPerPixel;
   mPixelWidth = mpScreen->format->BytesPerPixel;
   mIndexed = (mPixelWidth == 1);

   mLoss[0] = mpScreen->format->Rloss;
   mPos[0] = mpScreen->format->Rshift;
   mLoss[1] = mpScreen->format->Gloss;
   mPos[1] = mpScreen->format->Gshift;
   mLoss[2] = mpScreen->format->Bloss;
   mPos[2] = mpScreen->format->Bshift;

   unlockVideo();

   return 0;
}


void win32::closeVideo()
{
}


void win32::setColor(int index, int r, int g, int b, int value)
{
   System::setColor(index, r, g, b, value);

   if (-1 == value && mIndexed)
   {
      SDL_Color color;

      color.r = r;
      color.g = g;
      color.b = b;

      SDL_SetPalette(mpScreen, SDL_LOGPAL | SDL_PHYSPAL, &color, index, 1);
   }
}


void win32::lockVideo()
{
   System::lockVideo();

   if (SDL_MUSTLOCK(mpScreen))
      SDL_LockSurface(mpScreen);

   mWidth = mpScreen->w;
   mHeight = mpScreen->h;
   mPitch = mpScreen->pitch;
   mpPixels = (unsigned char *) mpScreen->pixels;
}


void win32::unlockVideo()
{
   System::unlockVideo();

   SDL_UnlockSurface(mpScreen);

   if (0 == mVideoLock)
      SDL_Flip(mpScreen);
}


int win32::pollEvents(eventHandler *pEventHandler)
{
   SDL_Event sdl_ev;
   
   while (SDL_PollEvent(&sdl_ev))
   {
      event ev;

      switch (sdl_ev.type)
      {
      case SDL_QUIT:
         return 1;
         break;

      case SDL_KEYDOWN:
         ev.type = EVENT_NONE;

         // TODO: hacktastic for now
         switch (sdl_ev.key.keysym.sym)
         {
         case SDLK_ESCAPE:
            ev.type = EVENT_QUIT;
            break;

         case SDLK_KP_PLUS:
            ev.type = EVENT_UNTHROTTLE;
            ev.data = 1;
            break;

         case SDLK_r:
            ev.type = EVENT_RESET;
            ev.data = 0;
            break;

         case SDLK_h:
            ev.type = EVENT_RESET;
            ev.data = 1;
            break;

         case SDLK_RETURN:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_START;
            break;

         case SDLK_TAB:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_SELECT;
            break;

         case SDLK_LCTRL:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_B;
            break;

         case SDLK_LALT:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_A;
            break;

         case SDLK_UP:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_UP;
            break;

         case SDLK_DOWN:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_DOWN;
            break;

         case SDLK_LEFT:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_LEFT;
            break;

         case SDLK_RIGHT:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_RIGHT;
            break;

         default:
            break;
         }
         pEventHandler->post(&ev);
         break;

      case SDL_KEYUP:
         // TODO: hacktastic for now
         ev.type = EVENT_NONE;

         switch (sdl_ev.key.keysym.sym)
         {
         case SDLK_KP_PLUS:
            ev.type = EVENT_UNTHROTTLE;
            ev.data = 0;
            break;

         case SDLK_RETURN:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_START;
            break;

         case SDLK_TAB:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_SELECT;
            break;

         case SDLK_LCTRL:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_B;
            break;

         case SDLK_LALT:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_A;
            break;

         case SDLK_UP:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_UP;
            break;

         case SDLK_DOWN:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_DOWN;
            break;

         case SDLK_LEFT:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_LEFT;
            break;

         case SDLK_RIGHT:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_RIGHT;
            break;

         default:
            break;
         }
         pEventHandler->post(&ev);
         break;

      default:
         break;
      }
   }

   return 0;
}


void win32::mark()
{
   mTimer = (long long) SDL_GetTicks() * 1000;
}


int win32::elapsed()
{
   long long now;
   int usecs;

   now = (long long) SDL_GetTicks() * 1000;
   usecs = (int) (now - mTimer);
   mTimer = now;

   return usecs;
}


void win32::usleep(int us)
{
   if (us > 0)
      SDL_Delay(us / 1000);
}
